from redis.client import StrictRedis

from redisopy.base.connection import DEFAULT_CONNECTION


class ConnMixin:
    """辅助数据库连接"""

    @property
    def conn(self) -> StrictRedis:
        """Get alive redis connection by connect"""
        return DEFAULT_CONNECTION[0]



from collections import UserList

from redisopy.containers.base import RedisBaseContainer


class RedisList(RedisBaseContainer, UserList):
    """
    用于操作 redis list，重写其所有可变序列方法。
    不支持切片赋值
        - s[i:j] = t
        - del s[i:j]
        - s[i:j:k] = t
        - del s[i:j:k]
    """

    def init_redis_value(self):
        if self.is_init:
            self.data = self.lrange()
        else:
            self.data = []

    def __setitem__(self, key, value):
        self.conn.conn.lset(self.key, key, value)
        self.data[key] = value

    def append(self, value: str):
        self.conn.conn.rpush(self.key, value)
        self.data.append(value)

    def extend(self, *elements) -> None:
        self.conn.conn.rpush(self.key, *elements)
        self.data.extend(elements)

    def insert(self, index: int, value: str):
        self.conn.conn.linsert(self.key, 'BEFORE', self.data[index], value)
        self.data.insert(index, value)

    def pop(self, i: int = -1):
        """only can pop last element"""
        value = self.conn.conn.rpop(self.key)
        self.data.pop(-1)
        return value

    def remove(self, item) -> None:
        self.conn.conn.lrem(self.key, 1, item)
        self.data.remove(item)

    def clear(self):
        self.conn.conn.delete(self.key)
        self.data.clear()

    def lrange(self):
        return self.conn.conn.lrange(self.key, 0, -1)

    def __str__(self):
        return f"{self.__class__.__name__}:{self.key}"

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return self.key == other.key

    def __hash__(self):
        return hash(self.key)



from redisopy.base.field import BaseField
from redisopy.errors import CannotOverwriteError
from redisopy.utils.model_var import ModelClassVar


class ModelMeta(type):

    def __new__(cls, name, bases, attrs):
        attrs["class_var"] = ModelClassVar(cls_name=name, meta=attrs.get("Meta", None))
        for key, value in attrs.items():
            if isinstance(value, BaseField):
                setattr(cls, key, value)
                attrs["class_var"].model_fields[key] = value
                attrs["class_var"].keys.add(key)
        return super().__new__(cls, name, bases, attrs)

    class Meta:
        # key_prefix
        pass


class BaseModel(metaclass=ModelMeta):
    """
    基础操作，主要是实例方法
    @property 标记的都是实例变量
    """
    class_var = None
    _record_id: int = None  # key 中的 id, 如果 model 有 id 字段，那么就是 id 字段的值。
    _key: str = ""  # redis instance key

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __str__(self):
        return f"{self.__class__.__name__}{str(self.fields)}"

    def __repr__(self):
        return str(self)

    def save(self, ex: int = 0, is_override: bool = False) -> int:
        if getattr(self, "id", None) and (not is_override):
            if self.class_var.conn.exists(self.key):
                raise CannotOverwriteError("Cannot overwrite the data in redis. If want, set is_override=True")
        self.class_var.conn.hset(self.key, mapping=self.redis_data)
        if ex:
            self.class_var.conn.expire(self.key, ex)
        return self.record_id

    def delete(self):
        self.class_var.conn.delete(self.key)

    @property
    def record_id(self):
        """如果 model 有 id 字段，那么就是 id 字段的值; 否则就通过 _record_id 来生成"""
        db_id = getattr(self, "id", None)
        return db_id if db_id else self._record_id

    @record_id.setter
    def record_id(self, value: int):
        self._record_id = value

    @property
    def key(self) -> str:
        if not self._key:
            prefix = self.class_var.key_prefix
            if self.record_id:
                self._key = f"{prefix}{self.record_id}"
            else:
                keys = self.class_var.conn.keys(f"{prefix}*")
                ids = [int(key.split(":")[-1]) for key in keys]
                self.record_id = max(ids) + 1 if ids else 1
                self._key = f"{prefix}{self.record_id}"
        return self._key

    @property
    def fields(self):
        data = {}
        for k in self.class_var.keys:
            data[k] = getattr(self, k)
        return data

    @property
    def redis_data(self):
        data = {}
        for k, v in self.class_var.model_fields.items():
            data[k] = v.validator.from_py_to_redis(getattr(self, k))
        return data



from redisopy.base.field import BaseField
from redisopy.utils.alias import ForkDatetime


class BooleanField(BaseField):
    field_type = bool


class StringField(BaseField):
    field_type = str


class IntField(BaseField):
    field_type = int


class FloatField(BaseField):
    field_type = float


class DatetimeField(BaseField):
    field_type = ForkDatetime



import inspect

import pendulum


class TypeValidator(object):
    """Validate/Convert that the value is of the correct type."""

    def __init__(self, t):
        self.type = t

    def from_redis_to_py(self, value):
        func_name = inspect.currentframe().f_code.co_name
        func = getattr(self, f'{func_name}_{self.type.__name__}')
        return func(value)

    def from_py_to_redis(self, value):
        func_name = inspect.currentframe().f_code.co_name
        func = getattr(self, f'{func_name}_{self.type.__name__}')
        return func(value)

    # 具体类型函数
    def from_py_to_redis_float(self, value):
        return str(value)

    def from_py_to_redis_int(self, value):
        return str(value)

    def from_py_to_redis_str(self, value):
        return str(value)

    def from_py_to_redis_bool(self, value):
        return '1' if value else ''

    def from_py_to_redis_fork_datetime(self, value):
        return value.in_timezone('UTC').format('YYYY-MM-DD HH:mm:ss')

    def from_redis_to_py_float(self, value):
        return float(value)

    def from_redis_to_py_int(self, value):
        return int(value)

    def from_redis_to_py_str(self, value):
        return str(value)

    def from_redis_to_py_bool(self, value):
        return bool(value)

    def from_redis_to_py_fork_datetime(self, value):
        if isinstance(value, pendulum.DateTime):
            return value.in_tz('UTC')
        return pendulum.parse(value, strict=False, tz='UTC')



from typing import Set

from redisopy.base.model import BaseModel
from redisopy.errors import NoSuchIdError


class Model(BaseModel):
    """增删改查, id 自动生成，不要加 id 字段"""

    @classmethod
    def filter(cls, include: Set[str] = None, exclude: Set[str] = None, **kwargs):
        """
        :param include/exclude: can use one
        :param kwargs: filter by field
        :return:
        """
        conn = cls.class_var.conn
        keys = conn.keys(f"{cls.class_var.key_prefix}*")
        res = []
        for key in keys:
            is_match = True
            if include:
                row = conn.hmget(key, include)
            elif exclude:
                sub_keys = cls.class_var.keys - exclude
                row = conn.hmget(key, sub_keys)
            else:
                row = conn.hgetall(key)

            for k, v in kwargs.items():
                if row[k] != v:
                    is_match = False
                    break
            if is_match:
                res.append(cls(**row))
        return res

    @classmethod
    def get_by_id(cls, r_id: int):
        conn = cls.class_var.conn
        key = f"{cls.class_var.key_prefix}{r_id}"
        row = conn.hgetall(key)
        if row:
            return cls(**row)
        else:
            raise NoSuchIdError(f"{cls.class_var.cls_name} No such id: {r_id}")

    def hset(self, key, value):
        self.class_var.conn.hset(self.key, key, value)
        setattr(self, key, value)

    def hincrease(self, key, value: int = 1):
        self.class_var.conn.hincrby(self.key, key, value)
        setattr(self, key, getattr(self, key) + value)

    def hincrease_float(self, key, value: float = 1.0):
        self.class_var.conn.hincrbyfloat(self.key, key, value)
        setattr(self, key, getattr(self, key) + value)

    def expire(self, ex: int):
        self.class_var.conn.expire(self.key, ex)



from abc import abstractmethod

from redis.exceptions import ResponseError

from redisopy.base.mixin import ConnMixin


class RedisBaseContainer:
    """"""
    conn = ConnMixin()

    def __init__(self, key: str, is_init: bool = False):
        """
        :param key:
        :param is_init: if not, not load data: Only update, not read
        """
        self.key = key
        self.is_init = is_init
        try:
            self.init_redis_value()
        except ResponseError as e:
            if "WRONGTYPE" in e.args[0]:
                raise TypeError(f"Key {self.key} is not a {self.__class__.__name__}")
            raise e

    @abstractmethod
    def init_redis_value(self):
        return NotImplemented



"""Use to connect redis."""
from redis.client import StrictRedis

DEFAULT_HOST = 'localhost'
DEFAULT_PORT = 6379
DEFAULT_DB = 0

DEFAULT_CONNECTION: list[StrictRedis] = []


def connect(host: str = DEFAULT_HOST, port: int = DEFAULT_PORT, db: int = DEFAULT_DB):
    """
    Use to connect redis.
    Exist a default link, can be overide.
    """
    if not DEFAULT_CONNECTION:
        redis = StrictRedis(host=host, port=port, db=db, decode_responses=True)
        DEFAULT_CONNECTION.append(redis)
    return DEFAULT_CONNECTION[0]



from collections.abc import MutableSet

from redisopy.containers.base import RedisBaseContainer


class RedisSet(RedisBaseContainer, MutableSet):
    data = set()
    """
    用于操作 redis set，重写其所有可变序列方法。
    """

    def init_redis_value(self):
        if self.is_init:
            self.data = self.smembers()
        else:
            self.data = set()

    def __contains__(self, item):
        return item in self.data

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def add(self, item):
        self.conn.conn.sadd(self.key, item)
        self.data.add(item)

    def discard(self, item):
        self.conn.conn.srem(self.key, item)
        self.data.discard(item)

    def clear(self):
        self.conn.conn.delete(self.key)
        self.data.clear()

    def smembers(self):
        return self.conn.conn.smembers(self.key)

    def __str__(self):
        return f"{self.__class__.__name__}:{self.key}"

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return self.data == other.data

    def __hash__(self):
        return hash(self.data)

    def __or__(self, other):
        return self.data | other.data

    def __and__(self, other):
        return self.data & other.data

    def __sub__(self, other):
        return self.data - other.data

    def __xor__(self, other):
        return self.data ^ other.data

    def __ior__(self, other):
        self.conn.conn.sunionstore(self.key, self.key, other.key)
        self.data |= other.data
        return self

    def __iand__(self, other):
        self.conn.conn.sinterstore(self.key, self.key, other.key)
        self.data &= other.data
        return self

    def __isub__(self, other):
        self.conn.conn.sdiffstore(self.key, self.key, other.key)
        self.data -= other.data
        return self

    def __ixor__(self, other):
        self.conn.conn.sunionstore(self.key, self.key, other.key)
        self.conn.conn.sinterstore(self.key, self.key, other.key)
        self.data ^= other.data
        return self

    def __le__(self, other):
        return self.data <= other.data

    def __lt__(self, other):
        return self.data < other.data

    def __ge__(self, other):
        return self.data >= other.data



from redisopy.base.connection import connect
from redisopy.containers.list import RedisList

connect(db=3)


def test_list():
    redis_list = RedisList('abc')
    # redis_list.clear()
    for i in 'abcde':
        redis_list.append(i)
    redis_list[0] = 'e'
    print(redis_list.data)

    redis_list.extend('f', 'g', 'h')
    redis_list.insert(0, 'a')
    var = redis_list.pop()
    print(var)



from redisopy.base.connection import connect
from redisopy.containers.set import RedisSet

connect(db=3)


def test_set():
    s = RedisSet('abcd')
    s.add('a')
    s.add('b')
    s.add('a')
    print('a' in s, 'b' in s, 'c' in s)



import pendulum

from redisopy.base.connection import connect
from redisopy.fields import StringField, IntField, BooleanField, DatetimeField
from redisopy.models import Model


class Person(Model):
    name = StringField()
    age = IntField()
    is_active = BooleanField()

    class Meta:
        key_prefix = "UpUp:person:"


class Person2(Model):
    name = StringField()
    age = IntField()

    class Meta:
        key_prefix = "UpUp:person2:"


connect(db=3)


def test_use_model():
    class PersonWithId(Model):
        id = IntField()
        name = StringField()
        age = IntField()

    person = Person(name='John', age=20, is_active=True)
    person.save()
    person2 = Person(name='John', age=20, is_active=False)
    person2.save()
    person3 = PersonWithId(name='John Has id', age=20, id=1)
    person3.save(is_override=True)
    rows = Person.filter()
    for r in rows:
        print(r)


def test_instance_delete():
    person = Person2(name='John', age=20)
    person.save()
    person.delete()


def test_get_instance():
    person = Person2(name='test', age=99)
    person.save()
    person2 = Person2.get_by_id(r_id=person.record_id)
    print(person2)


def test_time():
    class UserOnlinePeriods(Model):
        id = IntField()  # 用户id
        total = IntField()
        year = IntField()
        month = IntField()
        last_action_time = DatetimeField()

        class Meta:
            key_prefix = f"online:"

    user = UserOnlinePeriods(
        id=1, total=1, year=1, month=1, last_action_time=pendulum.now(tz='Asia/Shanghai')
    )
    user.save(is_override=True)
    print(user)



import setuptools

with open("README.md", "r", encoding='utf8') as fh:
    long_description = fh.read()

with open("requirements.txt", "r", encoding='utf8') as fh:
    requires = fh.read().replace('==', '>=')

setuptools.setup(
    name="redisopy",
    version="0.0.6",
    author="Steven Wang",
    author_email="brightstar8284@icloud.com",
    description="python redis orm",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/StevenLianaL/redisopy",
    packages=setuptools.find_packages(exclude=["tests*"]),
    install_requires=requires,
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
    ],
)



import random
from dataclasses import dataclass, field
from typing import Set, Any

from redisopy.base.mixin import ConnMixin


@dataclass
class ModelClassVar(ConnMixin):
    """Model class level var"""
    cls_name: str = ""
    meta: Any = None
    keys: Set[str] = field(default_factory=set)
    model_fields: dict = field(default_factory=dict)

    def __hash__(self):
        return hash(self.cls_name, random.randint(0, 100))

    def __eq__(self, other):
        return self.cls_name == other.cls_name

    @property
    def key_prefix(self) -> str:
        if self.meta:
            key_prefix = getattr(self.meta, "key_prefix", self.cls_name)
        else:
            key_prefix = f"{self.cls_name}:"
        return key_prefix



from redisopy.errors import TypeNotMatchError
from redisopy.utils.validation import TypeValidator


class FieldMeta(type):

    def __new__(cls, name, bases, attrs):
        attrs["validator"] = TypeValidator(attrs.get("field_type", None))
        return super().__new__(cls, name, bases, attrs)


class BaseField(metaclass=FieldMeta):
    _counter: int = 0
    field_type: type = None
    validator: TypeValidator = None

    def __init__(self, *, required=False, default=None, verbose_name='', help_text=''):
        self.required = required
        self.default = default
        self.verbose_name = verbose_name
        self.help_text = help_text

        cls = self.__class__
        prefix = cls.__name__
        index = cls._counter
        self.storage_name = f'{prefix}#{index}'
        cls._counter += 1

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return self.validator.from_redis_to_py(getattr(instance, self.storage_name, ''))

    def __set__(self, instance, value):
        try:
            setattr(instance, self.storage_name, self.validator.from_redis_to_py(value))
        except ValueError:
            raise TypeNotMatchError(f"Type not match. Expect {self.field_type}, but got {type(value)}")
